from celery.utils.log import get_task_logger
from storage import UploadedEntity

from portal import storage
from portal.async_celery import celery_app, submission_processor

log = get_task_logger(__name__)


@celery_app.task(name='upload-submission-to-storage')
def process_submission(new_submission_id: str):
    from portal.service.rest import RestService
    rest_service = RestService()
    new_submission = rest_service.find.submission(new_submission_id)
    project = new_submission.project
    course = project.course
    log.info(f"[SUBMIT] Processing submission: {new_submission.log_name}")
    if not project.config.test_files_commit_hash:
        log.warning(f"Project test files not found: {project.log_name}")
        update_project_test_files(course_id=course.id, project_id=project.id)
    new_submission = rest_service.find.submission(new_submission_id)
    processor = submission_processor.SubmissionProcessor(new_submission)
    processor.process_submission()


@celery_app.task(name='upload-results-to-storage')
def upload_results_to_storage(new_submission_id: str, path: str):
    path = str(path)
    from portal.service.rest import RestService
    rest_service = RestService()
    new_submission = rest_service.find.submission(new_submission_id)
    log.info(f"[SUBMIT] Processing results - upload to the storage for "
             f"{new_submission.log_name}: {path}")
    processor = submission_processor.SubmissionProcessor(new_submission)
    file_params = dict(source=dict(url=path, type='zip'))
    processor.upload_result(path=path, file_params=file_params)
    submission = processor.process_result()
    rest_service.submissions.write_entity(submission)


@celery_app.task(name='clone-submission-files')
def clone_submission_files(source_id: str, target_id: str):
    from portal.service.rest import RestService
    rest_service = RestService()
    source = rest_service.find.submission(source_id)
    target = rest_service.find.submission(target_id)
    log.info(f"[SUBMIT] Copy submission files: {source.log_name} -> {target.log_name}")
    processor = submission_processor.SubmissionProcessor(source)
    processor.clone(target)


@celery_app.task(name='start-processing-submission')
def start_processing_submission(submission_id: str, submission_params):
    from portal.service.rest import RestService
    rest_service = RestService()
    submission = rest_service.find.submission(submission_id)
    log.info(f"[SUBMIT] Processing submission - send to worker: {submission.log_name}")
    processor = submission_processor.SubmissionProcessor(submission, submission_params)
    processor.send_to_worker()


@celery_app.task(name='update-project-test-files')
def update_project_test_files(course_id: str, project_id: str):
    from portal.service.rest import RestService
    rest_service = RestService()

    course = rest_service.find.course(course_id)
    project = rest_service.find.project(course, project_id)
    log.info(f"[ASYNC] Updating test files for project: {project.log_name}")
    params = {
        'from_dir': project.config.test_files_subdir,
        'source': {
            'type': 'git',
            'url': project.config.test_files_source
        }
    }
    try:
        _update_test_files(rest_service, project, params)
    except Exception as ex:
        log.error(f"[ASYNC] Cannot update source files {ex}")


def _update_test_files(rest_service, project, params):
    updated_entity: UploadedEntity = storage.test_files.update(entity_id=project.id, **params)
    project.config.test_files_commit_hash = updated_entity.version
    log.debug(f"Updated project config {project.log_name}: {project.config}")
    rest_service.projects.write_entity(project.config)
