"""
Main Configuration module
Defines all the configuration types for the Flask Application
Types:
    - Development
    - Production
    - Test
"""

import datetime
import os
import tempfile

from portal import logger
from portal.tools import paths

log = logger.get_logger(__name__)


# pylint: disable=too-few-public-methods

class Config(object):
    """Base configuration
    """
    PROJECT_ROOT = paths.ROOT_DIR
    APP_DIR = paths.ROOT_DIR / 'portal'
    SECRET_KEY = os.environ.get('SECRET_KEY') or 'you-will-never-guess'
    JWT_SECRET_KEY = os.environ.get('JWT_SECRET_KEY') or 'super-safe'
    JWT_REFRESH_TOKEN_EXPIRES = datetime.timedelta(days=30)
    JWT_ACCESS_TOKEN_EXPIRES = datetime.timedelta(hours=1)
    API_PREFIX = os.environ.get('PORTAL_API_PREFIX', default="/api/v1.0/")
    CORS_SUPPORT_CREDENTIALS = True
    BROKER_URL = os.getenv('CELERY_BROKER_URL', 'redis://localhost:6379/0')
    CELERY_RESULT_BACKEND = os.getenv('CELERY_RESULT_BACKEND', BROKER_URL)
    SQLALCHEMY_DATABASE_URI = os.getenv('SQLALCHEMY_DATABASE_URI',
                                        f"sqlite:///{paths.ROOT_DIR}/devel.db")
    EMAIL_BACKEND = 'tests.utils.email_backend.EmailBackend'
    PORTAL_LOG_CONFIG = False
    PORTAL_DATA_WORKER_DOMAIN = os.getenv('PORTAL_DATA_WORKER_DOMAIN', 'localhost')
    LDAP_URL = os.getenv('LDAP_URL', None)
    GITLAB_BASE_DOMAIN = "gitlab.fi.muni.cz"
    GIT_REPO_BASE = os.getenv('GIT_REPO_BASE', f"git@{GITLAB_BASE_DOMAIN}")
    GITLAB_URL = os.getenv('GITLAB_URL', None)
    LOG_LEVEL_GLOBAL = os.getenv('LOG_LEVEL_GLOBAL', 'INFO')
    LOG_LEVEL_FILE = os.getenv('LOG_LEVEL_FILE', LOG_LEVEL_GLOBAL)
    LOG_LEVEL_CONSOLE = os.getenv('LOG_LEVEL_CONSOLE', LOG_LEVEL_GLOBAL)


class DevelopmentConfig(Config):
    """Development configuration
    """
    SQLALCHEMY_DATABASE_URI = os.getenv('SQLALCHEMY_DATABASE_URI',
                                        f"sqlite:///{paths.ROOT_DIR}/devel.db")
    SQLALCHEMY_TRACK_MODIFICATIONS = True
    PORTAL_ENV = 'dev'
    PORTAL_ADMIN_USER_PASSWORD = '789789'
    DEBUG = True
    PORTAL_STORAGE_BASE_DIR = f"{tempfile.gettempdir()}/kontr-portal"
    PORTAL_STORAGE_TEST_FILES_DIR = f"{PORTAL_STORAGE_BASE_DIR}/test-files"
    PORTAL_STORAGE_WORKSPACE_DIR = f"{PORTAL_STORAGE_BASE_DIR}/workspace"
    PORTAL_STORAGE_SUBMISSIONS_DIR = f"{PORTAL_STORAGE_BASE_DIR}/submissions"
    PORTAL_STORAGE_RESULTS_DIR = f"{PORTAL_STORAGE_BASE_DIR}/results"
    UPLOAD_FOLDER = f"{PORTAL_STORAGE_BASE_DIR}/upload"
    EMAIL_BACKEND = 'tests.utils.email_backend.EmailBackend'
    EMAIL_HOST = "localhost"
    EMAIL_PORT = "1025"
    EMAIL_DEFAULT_FROM = "kontr@example.com"
    PORTAL_LOG_CONFIG = True


class ProductionConfig(Config):
    """Production configuration
    """
    DEBUG = False
    LOG_LEVEL_GLOBAL = os.getenv('LOG_LEVEL_GLOBAL', 'INFO')
    LOG_LEVEL_FILE = os.getenv('LOG_LEVEL_FILE', LOG_LEVEL_GLOBAL)
    LOG_LEVEL_CONSOLE = os.getenv('LOG_LEVEL_CONSOLE', LOG_LEVEL_GLOBAL)


class TestConfig(Config):
    """Test configuration
    """
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    TESTING = True
    DEBUG = False
    SQLALCHEMY_DATABASE_URI = 'sqlite://'
    # SQLALCHEMY_ECHO = True
    PORTAL_ENV = 'dev'
    PORTAL_STORAGE_BASE_DIR = tempfile.mkdtemp()
    PORTAL_STORAGE_TEST_FILES_DIR = f"{PORTAL_STORAGE_BASE_DIR}/test-files"
    PORTAL_STORAGE_WORKSPACE_DIR = f"{PORTAL_STORAGE_BASE_DIR}/workspace"
    PORTAL_STORAGE_SUBMISSIONS_DIR = f"{PORTAL_STORAGE_BASE_DIR}/submissions"
    PORTAL_STORAGE_RESULTS_DIR = f"{PORTAL_STORAGE_BASE_DIR}/results"
    PORTAL_ADMIN_USER_USERNAME = 'admin'
    PORTAL_ADMIN_USER_PASSWORD = '789789'
    EMAIL_BACKEND = 'tests.utils.email_backend.EmailBackend'
    EMAIL_HOST = "localhost"
    EMAIL_PORT = "1025"
    EMAIL_DEFAULT_FROM = "kontr@example.com"
    CORS_ORIGINS = '*'
    BROKER_URL = 'redis://'
    CELERY_RESULT_BACKEND = BROKER_URL
    PORTAL_LOG_CONFIG = False
    GIT_REPO_BASE = os.getenv('GIT_REPO_BASE', f"git@gitlab.local")
    LOG_LEVEL_GLOBAL = 'DEBUG'
    LOG_LEVEL_FILE = LOG_LEVEL_GLOBAL
    LOG_LEVEL_CONSOLE = LOG_LEVEL_GLOBAL

# pylint: enable=too-few-public-methods


CONFIGURATIONS = dict(
    dev=DevelopmentConfig(),
    test=TestConfig(),
    production=ProductionConfig()
)
