"""
A collection of Mixins specifying common behaviour and attributes of database entities.
"""
import datetime

from sqlalchemy.ext.hybrid import hybrid_property

from portal import db
# maybe use server_default, server_onupdate instead of default, onupdate; crashes tests
# (https://stackoverflow.com/questions/13370317/sqlalchemy-default-datetime)
# pylint: disable=too-few-public-methods
from portal.tools.naming import sanitize_code_name


class EntityBase(object):
    """Entity mixin for the models
    Attributes:
        created_at(Column): Date when the entity has been created
        updated_at(Column): Date when the entity has been updated
    """
    EXCLUDED = []
    created_at = db.Column(db.TIMESTAMP, default=db.func.now())
    updated_at = db.Column(
        db.TIMESTAMP, default=db.func.now(), onupdate=db.func.now())


class NamedMixin(object):
    _name = db.Column('name', db.String(50), nullable=False)
    _codename = db.Column('codename', db.String(30), nullable=False)
    description = db.Column(db.Text)

    @hybrid_property
    def codename(self):
        return self._codename

    @codename.setter
    def codename(self, value: str):
        """Sets codename of the entity only the value is provided

        Codename setter will set codename only if value is provided.
        It will also set the name to the value, if the name is empty (None)
        Codename has to be sanitized - it means that it is slugyfied and shortened

        Args:
            value: Codename of the entity
        """
        if value:
            self._codename = sanitize_code_name(value)
            if self._name is None:
                self._name = self._codename

    @hybrid_property
    def name(self) -> str:
        """Gets name
        Returns (str): Name of the entity

        """
        return self._name

    @name.setter
    def name(self, value):
        """Sets the name of the entity
        Args:
            value(str): Name of the entity
        """
        if self._codename is None:
            self.codename = value
        self._name = value
# pylint: enable=too-few-public-methods
