"""
Logging configuration module
"""

import logging
from logging.config import dictConfig

import flask

from portal.tools import paths


class Logging:
    def __init__(self, app: flask.Flask = None):
        self._app = app

    @property
    def global_log_level(self):
        if not self._app:
            return 'INFO'
        return self._app.config.get('LOG_LEVEL_GLOBAL', 'INFO')

    @property
    def file_log_level(self):
        if not self._app:
            return 'INFO'
        return self._app.config.get('LOG_LEVEL_FILE', self.global_log_level)

    @property
    def console_log_level(self):
        if not self._app:
            return 'INFO'
        return self._app.config.get('LOG_LEVEL_CONSOLE', self.global_log_level)

    @property
    def handlers(self):
        return {
            'console': self.get_handler_console(),
            'portal_file': self.get_logger_file('portal'),
            'access_file': self.get_logger_file('access'),
            'auth_file': self.get_logger_file('auth'),
            'storage_file': self.get_logger_file('storage'),
            'flask_file': self.get_logger_file('flask')
        }

    @property
    def formatters(self):
        return {
            'verbose': {'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s'},

            'simple': {'format': '%(levelname)s %(message)s'},
            'colored_console': {
                '()': 'coloredlogs.ColoredFormatter',
                'format': "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                'datefmt': '%H:%M:%S'
            },
        }

    @property
    def loggers(self) -> dict:
        return {
            'portal': {
                'handlers': ['console', 'portal_file'],
                'level': self.global_log_level, 'propagate': True
            },
            'portal.auth_log': {
                'handlers': ['console', 'auth_file'],
                'level': self.global_log_level, 'propagate': True
            },
            'portal.access_log': {
                'handlers': ['console', 'access_file'],
                'level': self.global_log_level, 'propagate': True
            },
            'tests': {
                'handlers': ['console'],
                'level': self.global_log_level, 'propagate': True
            },
            'management': {
                'handlers': ['console'], 'level': self.global_log_level, 'propagate': True
            },
            'app': {
                'handlers': ['console'], 'level': self.global_log_level, 'propagate': True
            },
            'flask': {
                'handlers': ['console', 'flask_file'], 'level': self.global_log_level,
                'propagate': True
            },
            'werkzeug': {
                'handlers': ['console'], 'level': self.global_log_level, 'propagate': True
            },
            'storage': {
                'handlers': ['console', 'storage_file'], 'level': self.global_log_level,
                'propagate': True
            },
        }

    @property
    def config(self):
        return {
            'version': 1,
            'disable_existing_loggers': True,
            'formatters': self.formatters,
            'handlers': self.handlers,
            'loggers': self.loggers,
        }

    def get_logger_file(self, name, level: str = None):
        level = level or self.file_log_level
        return {
            'level': level,
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'verbose',
            'filename': str(paths.LOG_DIR / f'{name}.log'),
            'maxBytes': 5000000,  # 5MB
            'backupCount': 5
        }

    def get_handler_console(self, level: str = None):
        level = level or self.console_log_level
        return {
            'level': level, 'class': 'logging.StreamHandler', 'formatter': 'colored_console'
        }

    def load_config(self):
        """Loads config based on the config type
        Args:
        """
        add_custom_log_level()
        dictConfig(self.config)


TRACE_LOG_LVL = 9


def _trace(self, message, *args, **kws):
    if self.isEnabledFor(TRACE_LOG_LVL):
        # Yes, logger takes its '*args' as 'args'.
        self._log(TRACE_LOG_LVL, message, args, **kws)


def add_custom_log_level():
    logging.addLevelName(TRACE_LOG_LVL, 'TRACE')
    logging.Logger.trace = _trace


def get_logger(*args, **kwargs):
    logger = logging.getLogger(*args, **kwargs)
    return logger


def get_access_logger():
    return logging.getLogger('portal.access_log')


def get_auth_logger():
    return logging.getLogger('portal.auth_log')


ACCESS = get_access_logger()
AUTH = get_auth_logger()
