from typing import Union

from flask import Flask, Response, make_response, redirect, request, session
from flask_oauthlib.client import OAuth, OAuthRemoteApp
from flask_restplus import Namespace

from portal import logger, oauth
from portal.rest import rest_api
from portal.rest.custom_resource import CustomResource

log = logger.get_logger(__name__)

oauth_namespace = Namespace('oauth')


def extract_user_info(me: dict) -> dict:
    log.debug(f"[GITLAB] Received info: {me}")
    return dict(
        uco=None,
        name=me['name'],
        username=me['username'],
        email=me['email']
    )


def gitlab_enabled(app: Flask) -> bool:
    return app.config.get('GITLAB_URL') and \
           app.config.get('GITLAB_CLIENT_ID') and \
           app.config.get('GITLAB_CLIENT_SECRET')


def create_gitlab_app(oauth_app: OAuth) -> Union[OAuthRemoteApp, None]:
    if not gitlab_enabled(oauth_app.app):
        return None

    base_url = oauth_app.app.config['GITLAB_URL']
    client_id = oauth_app.app.config['GITLAB_CLIENT_ID']
    client_secret = oauth_app.app.config['GITLAB_CLIENT_SECRET']

    return oauth_app.remote_app(
        'gitlab',
        base_url=base_url,
        request_token_url=None,
        access_token_url=base_url + "/oauth/token",
        authorize_url=base_url + "/oauth/authorize",
        access_token_method='POST',
        consumer_key=client_id,
        consumer_secret=client_secret
    )


gitlab = create_gitlab_app(oauth_app=oauth)


@oauth_namespace.route('/login')
class OAuthLogin(CustomResource):
    def get(self):
        if not gitlab:
            return {'message': 'Gitlab OAuth is not enabled'}, 404

        callback = rest_api.url_for(OAuthLoginAuthorized, _external=True)

        log.debug(f"Callback set: {callback}")
        return gitlab.authorize(callback=callback)


@oauth_namespace.route('/login/authorized')
class OAuthLoginAuthorized(CustomResource):
    def get(self):
        if not gitlab:
            return {'message': 'Gitlab OAuth is not enabled'}, 404
        resp = gitlab.authorized_response()
        if resp is None:
            return 'Access denied: reason=%s error=%s' % (
                request.args['error'],
                request.args['error_description']
            )

        token = extract_token(resp)
        login = login_to_gitlab_with_token(token)
        log.debug(f"[GITLAB] Gitlab Login response: {login}")
        return login


@gitlab.tokengetter
def get_gitlab_oauth_token():
    token = session.get('gitlab_token')
    return token


def user_oauth_register(user_info):
    from portal.service.rest import RestService
    rest = RestService()
    new_user = rest.users.create(
        uco=None,
        username=user_info['username'],
        name=user_info['name'],
        email=user_info['email'],
        is_admin=False
    )
    log.info(f"[GITLAB] Created user={new_user}")
    return new_user


def user_login(user_info) -> Response:
    from portal.service.rest import RestService
    rest = RestService()
    user = rest.find.user(user_info['username'], throws=False)
    if not user:
        user = user_oauth_register(user_info)
    resp = make_response(redirect(oauth.app.config.get('FRONTEND_URL')))
    resp.set_cookie('username', user.username)
    return resp


def extract_token(resp):
    token = resp['access_token']
    log.debug(f"[GITLAB] Received token: {token}")
    session['gitlab_token'] = (token, '')
    return token


def login_to_gitlab_with_token(token):
    me = gitlab.get('/api/v4/user')
    user_info = extract_user_info(me.data)
    login = user_login(user_info)
    login.set_cookie('gitlab_token', token)
    return login
