from flask_jwt_extended import jwt_required
from flask_restplus import Namespace

from portal import logger, storage
from portal.rest import rest_helpers
from portal.rest.custom_resource import CustomResource
from portal.rest.schemas import SCHEMAS
from portal.tools.decorators import access_log

submissions_namespace = Namespace('submissions')

log = logger.get_logger(__name__)


@submissions_namespace.route('')
class SubmissionsResource(CustomResource):
    @jwt_required
    def get(self):
        submissions = self.rest.submissions.find_all()
        log.debug(f"[REST] Get submissions by {self.client.log_name}")
        return SCHEMAS.dump('submissions', submissions)


@submissions_namespace.route('/<string:sid>')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submission not found')
class SubmissionResource(CustomResource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submission found', model=submission_schema)
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        # Authorization
        log.debug(f"[REST] Get submission by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission(submission)
        return SCHEMAS.dump('submission', submission)

    @jwt_required
    @access_log
    @submissions_namespace.response(204, 'Submission deleted')
    def delete(self, sid: str):
        self.permissions.require.sysadmin()
        submission = self.find.submission(sid)
        log.info(f"[REST] Delete submission by {self.client.log_name}: {submission.log_name}")
        self.rest.submissions(submission).delete()
        return '', 204


@submissions_namespace.route('/<string:sid>/state')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionState(CustomResource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submission state', model=submission_state_schema)
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission state by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission(submission)
        return SCHEMAS.dump('submission', submission)

    @jwt_required
    @access_log
    @submissions_namespace.response(204, 'Submission state updated')
    def put(self, sid: str):
        client = self.rest.auth.client
        submission = self.find.submission(sid)
        self.permissions.require.sysadmin_or_self(submission.user.id)
        data = rest_helpers.parse_request_data(action='update_state', resource='submission_state')
        service = self.rest.submissions(submission)
        service.update_submission_state(client=client, data=data)
        log.info(f"[REST] Update submission state for {submission.log_name}  by "
                 f"{self.client.log_name}: {data}")
        return '', 204


@submissions_namespace.route('/<string:sid>/stats')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionStatistics(CustomResource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submission state', model=submission_state_schema)
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission statistics "
                  f"by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission(submission)
        return self.rest.submissions(submission).get_stats()


@submissions_namespace.route('/<string:sid>/files')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionFiles(CustomResource):
    @jwt_required
    def get(self, sid: str):
        raise NotImplementedError()


@submissions_namespace.route('/<string:sid>/files/sources/tree')
@submissions_namespace.doc({'sid': 'Submission id'})
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionSourcesTree(CustomResource):
    @jwt_required
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission sources tree"
                  f" by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission(submission)
        service = self.rest.storage(submission)
        return service.send_files_tree()


@submissions_namespace.route('/<string:sid>/files/sources')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionSourceFiles(CustomResource):
    @jwt_required
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission source files "
                  f"by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission(submission)
        service = self.rest.storage(submission)
        return service.send_file_or_zip()


@submissions_namespace.route('/<string:sid>/files/test_files/tree')
@submissions_namespace.doc({'sid': 'Submission id'})
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionTestFilesTree(CustomResource):
    @jwt_required
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission test files tree by "
                  f"{self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission_group(submission)
        service = self.rest.storage(project=submission.project)
        storage_entity = service.get_test_files_entity_from_storage()
        return service.send_files_tree(storage_entity)


@submissions_namespace.route('/<string:sid>/files/test_files')
@submissions_namespace.doc({'sid': 'Submission id'})
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionTestFiles(CustomResource):
    @jwt_required
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission test files"
                  f" by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission_group(submission)
        service = self.rest.storage(submission=submission)
        storage_entity = service.get_test_files_entity_from_storage()
        return service.send_file_or_zip(storage_entity)


@submissions_namespace.route('/<string:sid>/files/results/tree')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResultFilesTree(CustomResource):
    @jwt_required
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission result tree"
                  f" by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission_group(submission)
        storage_entity = storage.results.get(submission.id)
        service = self.rest.storage(submission)
        return service.send_files_tree(storage_entity)


@submissions_namespace.route('/<string:sid>/files/results')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResultFiles(CustomResource):
    @jwt_required
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission result files"
                  f" by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission_group(submission)
        storage_entity = storage.results.get(submission.id)
        service = self.rest.storage(submission=submission)
        return service.send_file_or_zip(storage_entity)

    @jwt_required
    @access_log
    def post(self, sid: str):
        submission = self.find.submission(sid)
        # authorization
        course = submission.project.course
        self.permissions(course=course).require.permissions(['evaluate_submissions'])
        # todo: authorize worker
        service = self.rest.submissions(submission)
        task = service.upload_results_to_storage()
        log.info(f"[REST] Upload submission results for {submission.log_name} by "
                 f"{self.client.log_name}")
        return {'new_task': task.id}


@submissions_namespace.route('/<string:sid>/resubmit')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionResubmit(CustomResource):
    @jwt_required
    @access_log
    # @submissions_namespace.response(201, 'New submission', model=submission_schema)
    def post(self, sid: str):
        source_submission = self.find.submission(sid)
        # authorization
        course = source_submission.project.course
        perm_service = self.permissions(course=course)
        perm_service.require.permissions(permissions=['resubmit_submissions'])

        data = rest_helpers.parse_request_data(action='resubmit', resource='submission')
        # create new submission by copying files from the source submission in
        # storage
        service = self.rest.submissions(source_submission)
        new_submission = service.copy_submission(note=data['note'])
        log.info(f"[REST] Resubmit submission {source_submission.log_name} by "
                 f"{self.client.log_name}: {data}")
        return SCHEMAS.dump('submission', new_submission), 201


@submissions_namespace.route('/<string:sid>/cancel')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
@submissions_namespace.response(204, 'Submissions canceled')
class SubmissionCancel(CustomResource):
    @jwt_required
    @access_log
    # @submissions_namespace.response(201, 'New submission', model=submission_schema)
    def post(self, sid: str):
        submission = self.find.submission(sid)
        # authorization
        course = submission.project.course
        perm_service = self.permissions(course=course)
        perm_service.require.sysadmin_or_self(submission.user.id)
        # create new submission by copying files from the source submission in
        # storage
        service = self.rest.submissions(submission)
        service.cancel_submission()
        log.info(f"[REST] Cancel submission {submission.log_name} by "
                 f"{self.client.log_name}")
        return '', 204


@submissions_namespace.route('/<string:sid>/review')
@submissions_namespace.param('sid', 'Submission id')
@submissions_namespace.response(404, 'Submissions not found')
class SubmissionReview(CustomResource):
    @jwt_required
    # @submissions_namespace.response(200, 'Submissions review', model=review_schema)
    def get(self, sid: str):
        submission = self.find.submission(sid)
        course = submission.project.course
        log.debug(f"[REST] Get submission review"
                  f" by {self.client.log_name}: {submission.log_name}")
        self.permissions(course=course).require.read_submission(submission)
        return SCHEMAS.dump('review', submission.review)

    @jwt_required
    @access_log
    # @submissions_namespace.response(201, 'Submissions review created', model=review_schema)
    def post(self, sid: str):
        client = self.rest.auth.client
        submission = self.find.submission(sid)
        course = submission.project.course
        perm_service = self.permissions(course=course)
        perm_service.require.write_review_for_submission(submission)
        data = rest_helpers.parse_request_data(action='create', resource='review')

        review_service = self.rest.reviews(submission=submission)
        if not submission.review:
            review_service.create()

        # also writes to db
        review_service.create_review_items(items=data['review_items'], author=client)
        log.info(f"[REST] Create submission review {submission.log_name} by "
                 f"{self.client.log_name}: {data}")
        return SCHEMAS.dump('review', submission.review), 201
