"""
Emails service
"""

import logging
from pathlib import Path

import yaml
from emails.template import JinjaTemplate
from flask_emails import Message

from portal.database.models import User
from portal.service import errors
from portal.tools import paths

log = logging.getLogger(__name__)

EMAILS_DIR = paths.RESOURCES_DIR / 'emails'


class EmailMessage:
    def __init__(self, subject: str, body: str, **kwargs):
        """Create email message
        Args:
            subject(str): Subject of the the mail
            body(str): Body of the mail
            **kwargs:
        """
        self._subject = JinjaTemplate(subject)
        self._body = JinjaTemplate(body)
        self._params = kwargs

    @property
    def subject(self) -> JinjaTemplate:
        """Gets subject of a mail
        Returns(str): Subject of the mail
        """
        return self._subject

    @property
    def body(self) -> JinjaTemplate:
        """Gets body content
        Returns: Email body content
        """
        return self._body

    def create_message(self) -> Message:
        """Creates new message

        Returns(Message): Message instance
        """
        message = Message(
            subject=self.subject,
            html=self.body,
            **self._params
        )
        return message


def _read_yml(full_path):
    with open(str(full_path), 'r') as stream:
        try:
            return yaml.load(stream)
        except yaml.YAMLError as exc:
            log.error(f"[EMAIL] Yaml parser has failed: {exc}")
            raise errors.PortalServiceError(f"Cannot load email template from {full_path}")


class EmailMessages:
    def __init__(self, email_service: 'EmailService'):
        self._emails = email_service

    @property
    def emails(self) -> 'EmailService':
        return self._emails

    def get(self, name) -> EmailMessage:
        return self.emails.create_message(name)

    def __getitem__(self, item):
        return self.get(name=item)

    @property
    def user_created(self):
        return self.get('user_created')

    @property
    def user_password_update(self):
        return self.get('user_password_update')

    @property
    def user_role_membership(self):
        return self.get('user_role_membership')


class EmailService:
    def __init__(self, service):
        self._rest_service = service
        self._messages = EmailMessages(self)

    @property
    def messages(self) -> EmailMessages:
        return self._messages

    @property
    def emails_dir(self) -> Path:
        return EMAILS_DIR

    def load_template(self, template: str, template_type='general') -> dict:
        full_path = self.emails_dir / template_type / f"{template}.yml"
        if not full_path.exists():
            log.error(f"[EMAIL] Emails template do not exists: {full_path}")
            raise errors.PortalServiceError(f"Cannot load email template from {full_path}")
        return _read_yml(full_path=full_path)

    def create_message(self, name, template_type='general'):
        template = self.load_template(name, template_type=template_type)
        return EmailMessage(**template)

    def notify_user(self, user: User, template: str, context: dict, **kwargs):
        context.update(kwargs)
        message = self.messages[template]
        if not message:
            log.error(f"[EMAILS] Cannot find template: {template}")
            raise errors.PortalServiceError(f"Cannot find template: {template}")
        email = user.email
        if email:
            to = (user.name, email)
            log.info(f"[EMAIL] Send message to user [{user.log_name} :: {email}]: {message}")
            message.create_message().send(to=to, render=context)
        return message


def notify_user(user: User, context: dict = None) -> Message:
    """Notifies the user
    Args:
        user(User): User instance
        collection(str): collection name
        action(str): action name
        context(dict): Message context

    Returns(Message): Sent Message instance
    """

    return send_message_to_user(user=user, message=message, context=context)
