import pytest

from portal.database.models import Review, ReviewItem, Submission, SubmissionState
from tests.rest.rest_tools import assert_response
from . import rest_tools


@pytest.fixture
def created_submission(rest_service, client) -> Submission:
    cpp = rest_service.find.course("testcourse1")
    p = cpp.projects[0]

    request_dict = {
        "project_params": {'tags': 'docker'},
        "file_params": {
            "source": {
                "type": "git",
                "url": "https://gitlab.fi.muni.cz/xkompis/test-hello-world.git",
                "branch": "master",
                "checkout": "master"
            }
        }
    }

    url = f"/courses/{cpp.codename}/projects/{p.name}/submissions"
    response = rest_tools.make_request(client, url, json=request_dict, method='post')
    assert_response(response, 201)
    new_submission = rest_tools.extract_data(response)
    return rest_service.find.submission(new_submission['id'])


def test_list_all_avail(client):
    db_subm = Submission.query.all()
    response = rest_tools.make_request(client, f'/submissions')
    assert_response(response, 200)

    resp_submissions = rest_tools.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


def test_list_all_avail_for_user(rest_service, client):
    student = rest_service.find.user('student1')
    db_subm = Submission.query.filter(Submission.user == student).all()
    response = rest_tools.make_request(client, f'/submissions?user=student1')
    assert_response(response, 200)

    resp_submissions = rest_tools.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


def test_list_all_avail_for_user_as_user(rest_service, client, student_credentials):
    student = rest_service.find.user('student1')
    db_subm = Submission.query.filter(Submission.user == student).all()
    response = rest_tools.make_request(client, f'/submissions', credentials=student_credentials)
    assert_response(response, 200)

    resp_submissions = rest_tools.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


def test_list_all_avail_for_user_as_teacher(rest_service, client, teacher1_credentials):
    response = rest_tools.make_request(client, f'/submissions', credentials=teacher1_credentials)
    assert_response(response, 200)

    resp_submissions = rest_tools.extract_data(response)
    for sub in resp_submissions:
        subm = rest_service.find.submission(sub['id'])
        print(subm.log_name)
    print([ role.log_name for role in rest_service.find.user('teacher1').roles])
    assert len(resp_submissions) == 5


def test_list_all_avail_for_course(rest_service, client):
    course = rest_service.find.course('testcourse1')
    db_subm = Submission.query.filter(Submission.course == course).all()
    response = rest_tools.make_request(client, f'/submissions?course=testcourse1', )
    assert_response(response, 200)
    resp_submissions = rest_tools.extract_data(response)
    num_of_submissions = len(resp_submissions)
    assert num_of_submissions == len(db_subm)
    assert num_of_submissions == 4


def test_list_all_avail_for_project(rest_service, client):
    course = rest_service.find.course('testcourse2')
    project = rest_service.find.project(course, 'hw01')
    db_subm = Submission.query.filter(Submission.project == project).all()
    response = rest_tools.make_request(client, f'/submissions?course=testcourse2&project=hw01')
    assert_response(response, 200)

    resp_submissions = rest_tools.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


# TODO: missing some tests (filter by group, role)
def test_list_all_avail_for_group(rest_service, client):
    course = rest_service.find.course('testcourse1')
    project = rest_service.find.project(course, 'hw01')
    group = rest_service.find.group(project.course, 'seminar01')
    db_subm = [sub for sub in Submission.query.filter(Submission.project == project).all() if
               group in sub.user.groups]
    response = rest_tools.make_request(client, f'/submissions?course=testcourse1&group=seminar01')
    assert_response(response, 200)
    resp_submissions = rest_tools.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


# missing tests for working with zip
def test_read(client):
    submissions = Submission.query.all()
    submission = submissions[0]
    response = rest_tools.make_request(client, f'/submissions/{submission.id}', )
    assert_response(response, 200)

    submission = rest_tools.extract_data(response)
    rest_tools.assert_submission_in(submissions, submission)


def test_student_is_able_to_access_own_submission(client, rest_service, student_credentials):
    student = rest_service.find.user('student1')
    submission = student.submissions[0]
    response = rest_tools.make_request(client, f'/submissions/{submission.id}',
                                       credentials=student_credentials)

    assert_response(response, 200)


def test_delete(client):
    submissions = Submission.query.all()
    submissions_len = len(submissions)
    s = submissions[0]
    assert s.review
    reviews_len = len(Review.query.all())
    response = rest_tools.make_request(client, f'/submissions/{s.id}', method='delete')
    assert_response(response, 204)
    submissions_updated = Submission.query.all()
    assert len(submissions_updated) == submissions_len - 1
    assert len(Review.query.all()) == reviews_len - 1


def test_read_state(client):
    submissions = Submission.query.all()
    s = submissions[0]
    response = rest_tools.make_request(
        client, f'/submissions/{s.id}/state', )
    assert_response(response, 200)

    state = rest_tools.extract_data(response)['state']
    assert state == s.state.name


# needs more tests for state transitions
def test_update_state(client):
    submissions = Submission.query.all()
    s = submissions[0]
    request_dict = {"state": "CANCELLED"}

    response = rest_tools.make_request(client, f'/submissions/{s.id}/state', json=request_dict,
                                       method='put')
    assert_response(response, 204)
    s_updated = Submission.query.all()[0]
    assert s_updated.state == SubmissionState.CANCELLED


# does not pass - needs a Storage mock or config
def test_resubmit(client, created_submission):
    submission = created_submission

    request_dict = {"note": "resubmit reason"}
    response = rest_tools.make_request(client, f'/submissions/{submission.id}/resubmit',
                                       json=request_dict, method='post')

    assert_response(response, 201)
    submission = rest_tools.extract_data(response)

    submissions_updated = Submission.query.all()
    rest_tools.assert_submission_in(submissions_updated, submission)


def test_read_review(client):
    submissions = Submission.query.all()
    s = submissions[0]
    assert s.review

    response = rest_tools.make_request(client, f'/submissions/{s.id}/review')
    assert_response(response, 200)

    review = rest_tools.extract_data(response)
    rest_tools.assert_review(expected=s.review, actual=review)


def test_create_review(client):
    s = Submission.query.filter_by(state=SubmissionState.IN_PROGRESS).first()
    assert s
    len_reviews = len(Review.query.all())
    len_review_items = len(ReviewItem.query.all())

    request_dict = {
        "review_items": [
            {
                "file": "src/main.c",
                "line": 10,
                "content": "This is wrong!"
            },
            {
                "file": "src.main.c",
                "line": 15,
                "content": "Nice job!"
            }
        ]
    }

    response = rest_tools.make_request(client, f'/submissions/{s.id}/review', json=request_dict,
                                       method='post')

    assert_response(response, 201)

    review = rest_tools.extract_data(response)
    s_updated = Submission.query.filter_by(project_id=s.project_id,
                                           state=SubmissionState.IN_PROGRESS).first()
    assert len(Review.query.all()) == len_reviews + 1
    assert len(ReviewItem.query.all()) == len_review_items + 2
    rest_tools.assert_review(s_updated.review, review)
